#!/usr/bin/python
# $Id: server.py,v 1.3 2003/06/15 13:04:29 peter Exp $

""" This module provides a simple server real IGS clients can connect to using
localhost:8888. Additionally for a test and debugging run the client of this
demonstration should connect to this server to allow some basic control.
To bind the server to another port than the default 8888 use the commandline
parameter -b <port>."""


import socket, sys, threading, string, __init__, IGSCP.__init__
from igs_messages import *

# List of threads containing a tuple (thread, connection) to store data of each connected client
threads = []

# Global socket variable
s = None

#
# This function is called within an own thread to handle communication with one client
#
def client_conn(conn, addr):
    while 1:
        try:
            data = conn.recv(1024)
        except KeyboardInterrupt:  # Catch Ctrl-C
            shutdown()
            return
        except:
            return

        # Client disconnected
        if not data or not conn:
            print 'Thread ended.'
            return

        # Strip trailing linebreak
        data = string.rstrip(data)
        print 'Recieved:', data

        # Strip leading "chatter" and "tell" commands
        if data[0:7] == 'chatter':
            data = data[8:len(data)]
        elif data[0:4] == 'tell':
            data = data[5:len(data)]

        # Server control commands
        if data[0:2] == 'SC' or data[0:2] == 'sc':
            command = data[3:len(data)]
            print 'Command: %s' % command

            if command == 'shutdown':
                shutdown()
                return
            elif command == 'list clients':
                for t, c in threads:
                    print t, c
            elif command == 'request version' or command == 'rv':
                broadcast('#INFO:REQUEST:VERSION', conn)
            elif command == 'request client' or command == 'rc':
                broadcast('#INFO:REQUEST:CLIENT', conn)
            elif command == 'request features' or command == 'rf':
                broadcast('#INFO:REQUEST:FEATURES', conn)
            elif command == 'request photo' or command == 'rp':
                broadcast('#INFO:REQUEST:PHOTO', conn)
            elif command[0:8] == 'add mark':
                broadcast('#MARK:ADD %s' % command[9:len(command)], conn)
            elif command[0:2] == 'am':
                broadcast('#MARK:ADD %s' % command[3:len(command)], conn)
            elif command[0:11] == 'remove mark':
                broadcast('#MARK:REMOVE %s' % command[12:len(command)], conn)
            elif command[0:2] == 'rm':
                broadcast('#MARK:REMOVE %s' % command[3:len(command)], conn)
            elif command[0:9] == 'move mark':
                broadcast('#MARK:MOVE %s' % command[10:len(command)], conn)
            elif command[0:2] == 'mm':
                broadcast('#MARK:MOVE %s' % command[3:len(command)], conn)
            else:
                send(conn, 'Unknown server control command.')
        # Client disconnecting
        elif data == 'quit' or data == 'exit' or data == 'bye':
            send(conn, 'bye')
            print 'Thread ended.'
            return
        # IGS fake commands
        elif data == 'teach 19':
            send(conn, IGS_FAKE_TEACH)
        elif data == 'status 91':
            send(conn, IGS_FAKE_STATUS)
        # IGSCP command. Forward them to all other clients
        elif data[0] == '#':
            broadcast(data, conn)
        # Unknown client command
        else:
            send(conn, 'Unknown command.')

def send(conn, text):
    # gGo workaround, add whitespace and linebreak. This should be a commandline parameter.
    text += ' \n'

    conn.send(text)

#
# Send text to all connected clients except the requesting one if the req argument is given.
# If req is None, send to all clients without exception
#
def broadcast(text, req=None):
    print 'Broadcasting: %s' % text

    # gGo workaround, add whitespace and linebreak. This should be a commandline parameter.
    text += ' \n'

    for t, c in threads:
        if t.isAlive() and (not req or req != c):
            print 'Sending ' + text + ' to', c
            c.send(text)

#
# Shutdown server and inform clients
# This sometimes still hangs because one thread is locked in the recieve command,
# but a forceful Ctrl-C in the server terminal usually does the trick.
#
def shutdown():
    global s

    print 'Shutting down server...'

    # Inform connected clients
    for t, c in threads:
        if t.isAlive():
            print 'Sending bye to', c
            c.send('Shutting down server')
            try:
                c.shutdown(2)
            except socket.error:
                pass
            c.close()

    try:
        s.shutdown(2)
    except socket.error:
        pass
    s.close()
    sys.exit(0)

#
# Remove dead threads from threads list
#
def cleanup_threads():
    for t, c in threads:
        if not t.isAlive():
            c.shutdown(2)
            c.close()
            threads.remove((t, c))

#
# Run server and wait for connections
#
def run_server(port=None):
    global s

    if not port:
        port = __init__.PORT

    # Create socket and listen on HOST:PORT
    try:
        s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        s.bind((__init__.HOST, port))
        s.listen(1)
    except socket.error:
        print 'Failed to create socket on %s:%d.\nThis port might be already in use.' % (__init__.HOST, port)
        sys.exit(1)

    print 'Server started on port %d\nUse Ctrl-C to shutdown' % port

    # Server loop
    while 1:
        # Wait for client connection
        print '\nWaiting for connection'
        try:
            conn, addr = s.accept()
        except KeyboardInterrupt:  # Catch Ctrl-C
            # Shutdown server
            shutdown()
            break
        except socket.error:
            break
        print 'Connected from: ', addr

        # Remove dead threads from list
        cleanup_threads()

        # Create thread for client connection loop
        t = threading.Thread(target=client_conn, args=(conn, addr))
        threads.append((t, conn))
        t.start()
        print 'Thread created.'

#
# Server main method. Parse commandline parameters and then start the server.
#
def main(args=None):
    import commandline
    result = None
    if args and len(args) > 1:
        result = commandline.parse_commandline(args[1:len(args)])
        if not result:
            sys.exit(0)

    print 'IGSCPDemo server %s using IGSCP version %s\n' % (__init__.VERSION, IGSCP.__init__.IGSCP_VERSION)
    run_server(result)

if __name__ == '__main__':
    main(sys.argv)
